
/*global __DEV__*/
const __DEV__ = true;
'use strict';

const isDev = typeof __DEV__ === 'boolean' && __DEV__;
import PropTypeLocationNames from '../validation/PropTypeLocationNames';
let PropTypesSecret = 'SECRET_DO_NOT_PASS_THIS_OR_YOU_WILL_BE_FIRED';

import emptyFunction from './emptyFunction';
import warning from './warning';

let ANONYMOUS = '<<anonymous>>';

let ReactPropTypes = {
  array: createPrimitiveTypeChecker('array'),
  bool: createPrimitiveTypeChecker('boolean'),
  func: createPrimitiveTypeChecker('function'),
  number: createPrimitiveTypeChecker('number'),
  object: createPrimitiveTypeChecker('object'),
  string: createPrimitiveTypeChecker('string'),
  symbol: createPrimitiveTypeChecker('symbol'),

  any: createAnyTypeChecker(),
  arrayOf: createArrayOfTypeChecker,
  instanceOf: createInstanceTypeChecker,
  objectOf: createObjectOfTypeChecker,
  oneOf: createEnumTypeChecker,
  oneOfType: createUnionTypeChecker,
  shape: createShapeTypeChecker,
};

/*eslint-disable no-self-compare*/
function is(x, y) {
  if (x === y) {
    return x !== 0 || 1 / x === 1 / y;
  } else {
    return x !== x && y !== y;
  }
}
function PropTypeError(message) {
  this.message = message;
  this.stack = '';
}
PropTypeError.prototype = Error.prototype;

function createChainableTypeChecker(validate) {
  let manualPropTypeCallCache;
  if (isDev) {
    manualPropTypeCallCache = {};
  }
  function checkType(
    isRequired,
    props,
    propName,
    componentName,
    location,
    propFullName,
    secret
  ) {
    componentName = componentName || ANONYMOUS;
    propFullName = propFullName || propName;
    if (isDev) {
      if (
        secret !== PropTypesSecret &&
        typeof console !== 'undefined'
      ) {
        let cacheKey = `${componentName}:${propName}`;
        if (!manualPropTypeCallCache[cacheKey]) {
          warning(
            false,
            'You are manually calling a React.PropTypes validation ' +
            'function for the `%s` prop on `%s`. This is deprecated ' +
            'and will not work in the next major version. You may be ' +
            'seeing this warning due to a third-party PropTypes library. ' +
            'See https://fb.me/react-warning-dont-call-proptypes for details.',
            propFullName,
            componentName
          );
          manualPropTypeCallCache[cacheKey] = true;
        }
      }
    }
    if (props[propName] == null) {
      let locationName = PropTypeLocationNames[location];
      if (isRequired) {
        if (props[propName] === null) {
          return new PropTypeError(
            `The ${locationName} \`${propFullName}\` is marked as required ` +
            `in \`${componentName}\`, but its value is \`null\`.`
          );
        }
        return new PropTypeError(
          `The ${locationName} \`${propFullName}\` is marked as required in ` +
          `\`${componentName}\`, but its value is \`undefined\`.`
        );
      }
      return null;
    } else {
      return validate(
        props,
        propName,
        componentName,
        location,
        propFullName,
      );
    }
  }

  let chainedCheckType = checkType.bind(null, false);
  chainedCheckType.isRequired = checkType.bind(null, true);

  return chainedCheckType;
}

function createPrimitiveTypeChecker(expectedType) {
  function validate(
    props,
    propName,
    componentName,
    location,
    propFullName,
    secret
  ) {
    let propValue = props[propName];
    let propType = getPropType(propValue);
    if (propType !== expectedType) {
      let locationName = PropTypeLocationNames[location];
      // `propValue` being instance of, say, date/regexp, pass the 'object'
      // check, but we can offer a more precise error message here rather than
      // 'of type `object`'.
      let preciseType = getPreciseType(propValue);

      return new PropTypeError(
        `Invalid ${locationName} \`${propFullName}\` of type ` +
        `\`${preciseType}\` supplied to \`${componentName}\`, expected ` +
        `\`${expectedType}\`.`
      );
    }
    return null;
  }
  return createChainableTypeChecker(validate);
}

function createAnyTypeChecker() {
  return createChainableTypeChecker(emptyFunction.thatReturns(null));
}

function createArrayOfTypeChecker(typeChecker) {
  function validate(props, propName, componentName, location, propFullName) {
    if (typeof typeChecker !== 'function') {
      return new PropTypeError(
        `Property \`${propFullName}\` of component \`${componentName}\` has invalid PropType notation inside arrayOf.`
      );
    }
    let propValue = props[propName];
    if (!Array.isArray(propValue)) {
      let locationName = PropTypeLocationNames[location];
      let propType = getPropType(propValue);
      return new PropTypeError(
        `Invalid ${locationName} \`${propFullName}\` of type ` +
        `\`${propType}\` supplied to \`${componentName}\`, expected an array.`
      );
    }
    for (let i = 0; i < propValue.length; i++) {
      let error = typeChecker(
        propValue,
        i,
        componentName,
        location,
        `${propFullName}[${i}]`,
        PropTypesSecret
      );
      if (error instanceof Error) {
        return error;
      }
    }
    return null;
  }
  return createChainableTypeChecker(validate);
}

function createInstanceTypeChecker(expectedClass) {
  function validate(props, propName, componentName, location, propFullName) {
    if (!(props[propName] instanceof expectedClass)) {
      let locationName = PropTypeLocationNames[location];
      let expectedClassName = expectedClass.name || ANONYMOUS;
      let actualClassName = getClassName(props[propName]);
      return new PropTypeError(
        `Invalid ${locationName} \`${propFullName}\` of type ` +
        `\`${actualClassName}\` supplied to \`${componentName}\`, expected ` +
        `instance of \`${expectedClassName}\`.`
      );
    }
    return null;
  }
  return createChainableTypeChecker(validate);
}

function createEnumTypeChecker(expectedValues) {
  if (!Array.isArray(expectedValues)) {
    warning(false, 'Invalid argument supplied to oneOf, expected an instance of array.');
    return emptyFunction.thatReturnsNull;
  }

  function validate(props, propName, componentName, location, propFullName) {
    let propValue = props[propName];
    for (let i = 0; i < expectedValues.length; i++) {
      if (is(propValue, expectedValues[i])) {
        return null;
      }
    }

    let locationName = PropTypeLocationNames[location];
    let valuesString = JSON.stringify(expectedValues);
    return new PropTypeError(
      `Invalid ${locationName} \`${propFullName}\` of value \`${propValue}\` ` +
      `supplied to \`${componentName}\`, expected one of ${valuesString}.`
    );
  }
  return createChainableTypeChecker(validate);
}

function createObjectOfTypeChecker(typeChecker) {
  function validate(props, propName, componentName, location, propFullName) {
    if (typeof typeChecker !== 'function') {
      return new PropTypeError(
        `Property \`${propFullName}\` of component \`${componentName}\` has invalid PropType notation inside objectOf.`
      );
    }
    let propValue = props[propName];
    let propType = getPropType(propValue);
    if (propType !== 'object') {
      let locationName = PropTypeLocationNames[location];
      return new PropTypeError(
        `Invalid ${locationName} \`${propFullName}\` of type ` +
        `\`${propType}\` supplied to \`${componentName}\`, expected an object.`
      );
    }
    for (let key in propValue) {
      if (propValue.hasOwnProperty(key)) {
        let error = typeChecker(
          propValue,
          key,
          componentName,
          location,
          `${propFullName}.${key}`,
          PropTypesSecret
        );
        if (error instanceof Error) {
          return error;
        }
      }
    }
    return null;
  }
  return createChainableTypeChecker(validate);
}

function createUnionTypeChecker(arrayOfTypeCheckers) {
  if (!Array.isArray(arrayOfTypeCheckers)) {
    warning(false, 'Invalid argument supplied to oneOfType, expected an instance of array.');
    return emptyFunction.thatReturnsNull;
  }

  function validate(props, propName, componentName, location, propFullName) {
    for (let i = 0; i < arrayOfTypeCheckers.length; i++) {
      let checker = arrayOfTypeCheckers[i];
      if (
        checker(
          props,
          propName,
          componentName,
          location,
          propFullName,
          PropTypesSecret
        ) == null
      ) {
        return null;
      }
    }

    let locationName = PropTypeLocationNames[location];
    return new PropTypeError(
      `Invalid ${locationName} \`${propFullName}\` supplied to ` +
      `\`${componentName}\`.`
    );
  }
  return createChainableTypeChecker(validate);
}

function createShapeTypeChecker(shapeTypes) {
  function validate(props, propName, componentName, location, propFullName) {
    let propValue = props[propName];
    let propType = getPropType(propValue);
    if (propType !== 'object') {
      let locationName = PropTypeLocationNames[location];
      return new PropTypeError(
        `Invalid ${locationName} \`${propFullName}\` of type \`${propType}\` ` +
        `supplied to \`${componentName}\`, expected \`object\`.`
      );
    }
    for (let key in shapeTypes) {
      let checker = shapeTypes[key];
      if (!checker) {
        continue;
      }
      let error = checker(
        propValue,
        key,
        componentName,
        location,
        `${propFullName}.${key}`,
        PropTypesSecret
      );
      if (error) {
        return error;
      }
    }
    return null;
  }
  return createChainableTypeChecker(validate);
}

function isSymbol(propType, propValue) {
  // Native Symbol.
  if (propType === 'symbol') {
    return true;
  }

  // 19.4.3.5 Symbol.prototype[@@toStringTag] === 'Symbol'
  if (propValue['@@toStringTag'] === 'Symbol') {
    return true;
  }

  // Fallback for non-spec compliant Symbols which are polyfilled.
  if (typeof Symbol === 'function' && propValue instanceof Symbol) {
    return true;
  }

  return false;
}

// Equivalent of `typeof` but with special handling for array and regexp.
function getPropType(propValue) {
  let propType = typeof propValue;
  if (Array.isArray(propValue)) {
    return 'array';
  }
  if (propValue instanceof RegExp) {
    // Old webkits (at least until Android 4.0) return 'function' rather than
    // 'object' for typeof a RegExp. We'll normalize this here so that /bla/
    // passes PropTypes.object.
    return 'object';
  }
  if (isSymbol(propType, propValue)) {
    return 'symbol';
  }
  return propType;
}

// This handles more types than `getPropType`. Only used for error messages.
// See `createPrimitiveTypeChecker`.
function getPreciseType(propValue) {
  let propType = getPropType(propValue);
  if (propType === 'object') {
    if (propValue instanceof Date) {
      return 'date';
    } else if (propValue instanceof RegExp) {
      return 'regexp';
    }
  }
  return propType;
}

// Returns class name of the object, if any.
function getClassName(propValue) {
  if (!propValue.constructor || !propValue.constructor.name) {
    return ANONYMOUS;
  }
  return propValue.constructor.name;
}

module.exports = ReactPropTypes;
